<?php

namespace Drupal\redirect4xx\EventSubscriber;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Cache\CacheableRedirectResponse;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\EventSubscriber\HttpExceptionSubscriberBase;
use Drupal\Core\Routing\AdminContext;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Symfony\Component\HttpKernel\Event\ExceptionEvent;

/**
 * Redirect 403 to User Login event subscriber.
 */
class Redirect4xxSubscriber extends HttpExceptionSubscriberBase {

  /**
   * Constructs a new Redirect4xxSubscriber.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The configuration factory.
   * @param \Drupal\Core\Routing\AdminContext $adminContext
   *   The admin context.
   * @param \Drupal\Core\Session\AccountInterface $currentUser
   *   The current user.
   */
  public function __construct(
    protected ConfigFactoryInterface $configFactory,
    protected AdminContext $adminContext,
    protected AccountInterface $currentUser,
  ) {}

  /**
   * {@inheritdoc}
   */
  protected function getHandledFormats() {
    return ['html'];
  }

  /**
   * Generates cacheable metadata for the 4xx redirection.
   *
   * This method ensures that the redirection response is properly cached
   * while remaining up-to-date when necessary. It includes:
   * - The original exception's cache metadata.
   * - A custom cache tag ('4xx-response') to track 4xx-related redirects.
   * - Configuration dependencies, ensuring cache is invalidated if settings
   *   change.
   *
   * @param \Symfony\Component\HttpKernel\Event\ExceptionEvent $event
   *   The exception event that triggered the redirection.
   *
   * @return \Drupal\Core\Cache\CacheableMetadata
   *   A cacheable metadata object containing dependencies and cache tags.
   */
  protected function getCacheableMetadata(ExceptionEvent $event) {
    // Add caching dependencies so the cache of the redirection will be
    // updated when necessary.
    $cacheable_metadata = new CacheableMetadata();
    // Add original response cache metadata.
    $cacheable_metadata->addCacheableDependency($event->getThrowable());
    // We still need to add the client error tag manually since the core
    // wil not recognize our redirection as an error.
    $cacheable_metadata->addCacheTags(['4xx-response']);
    // Add our config cache metadata.
    $config = $this->configFactory->get('redirect4xx.settings');
    $cacheable_metadata->addCacheableDependency($config);

    return $cacheable_metadata;
  }

  /**
   * Redirects on 4xx kernel exceptions.
   *
   * @param \Symfony\Component\HttpKernel\Event\ExceptionEvent $event
   *   The Event to process.
   */
  public function on4xx(ExceptionEvent $event) {
    /** @var \Symfony\Component\HttpKernel\Exception */
    $http_exception = $event->getThrowable();
    $status_code = $http_exception->getStatusCode();

    $redirects = $this->configFactory
      ->get('redirect4xx.settings')
      ->get('redirects');

    if (isset($redirects[$status_code])) {
      $redirect = $redirects[$status_code];
    }
    elseif (isset($redirects['4xx'])) {
      $redirect = $redirects['4xx'];
    }
    else {
      return;
    }

    // Prevent infinite redirect loops.
    $current_path = $event->getRequest()->getPathInfo();
    if ($current_path == $redirect['destination_path']) {
      return;
    }

    $enable_on_admin_route = !empty($redirect['enable_on_admin_route']);
    if (!$enable_on_admin_route && $this->adminContext->isAdminRoute()) {
      return;
    }

    $redirect_authenticated_users = !empty($redirect['redirect_authenticated_users']);
    if (!$redirect_authenticated_users && $this->currentUser->isAuthenticated()) {
      return;
    }

    $options = ['absolute' => TRUE];
    $url = Url::fromUserInput($redirect['destination_path'], $options)->toString();

    $response = new CacheableRedirectResponse($url, $redirect['redirect_status_code']);
    // Attach cache metadata to the response.
    $cacheable_metadata = $this->getCacheableMetadata($event);
    $response->addCacheableDependency($cacheable_metadata);

    $event->setResponse($response);
  }

}
