<?php

namespace Drupal\redirect4xx\Form;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;

/**
 * Configure Redirect 4xx settings for this site.
 *
 * @internal
 */
class SettingsForm extends ConfigFormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'redirect4xx_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['redirect4xx.settings'];
  }

  /**
   * Returns available redirect status codes.
   *
   * @return array
   *   An associative array of redirect status codes and their descriptions.
   */
  protected function getRedirectStatusOptions(): array {
    return [
      '301' => $this->t('301 Moved Permanently'),
      '302' => $this->t('302 Found'),
      '303' => $this->t('303 See Other'),
      '307' => $this->t('307 Temporary Redirect'),
      '308' => $this->t('308 Permanent Redirect'),
    ];
  }

  /**
   * Returns the list of 4xx client error status codes.
   *
   * @return array
   *   An associative array of HTTP 4xx error codes and their descriptions.
   */
  protected function getErrorStatusCodes(): array {
    return [
      '400' => $this->t('400 Bad Request'),
      '401' => $this->t('401 Unauthorized'),
      '402' => $this->t('402 Payment Required'),
      '403' => $this->t('403 Forbidden'),
      '404' => $this->t('404 Not Found'),
      '405' => $this->t('405 Method Not Allowed'),
      '406' => $this->t('406 Not Acceptable'),
      '407' => $this->t('407 Proxy Authentication Required'),
      '408' => $this->t('408 Request Timeout'),
      '409' => $this->t('409 Conflict'),
      '410' => $this->t('410 Gone'),
      '411' => $this->t('411 Length Required'),
      '412' => $this->t('412 Precondition Failed'),
      '413' => $this->t('413 Payload Too Large'),
      '414' => $this->t('414 URI Too Long'),
      '415' => $this->t('415 Unsupported Media Type'),
      '416' => $this->t('416 Range Not Satisfiable'),
      '417' => $this->t('417 Expectation Failed'),
      '418' => $this->t("418 I'm a teapot"),
      '421' => $this->t('421 Misdirected Request'),
      '422' => $this->t('422 Unprocessable Entity'),
      '423' => $this->t('423 Locked'),
      '424' => $this->t('424 Failed Dependency'),
      '425' => $this->t('425 Too Early'),
      '426' => $this->t('426 Upgrade Required'),
      '428' => $this->t('428 Precondition Required'),
      '429' => $this->t('429 Too Many Requests'),
      '431' => $this->t('431 Request Header Fields Too Large'),
      '451' => $this->t('451 Unavailable For Legal Reasons'),
      '4xx' => $this->t('4xx Fallback'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $redirects = $form_state->getValue('redirects');
    if (!isset($redirects)) {
      $config = $this->config('redirect4xx.settings');
      $redirects = $config->get('redirects') ?? [];
    }
    ksort($redirects);

    $status_codes = $this->getErrorStatusCodes();
    $redirect_status_options = $this->getRedirectStatusOptions();

    // Remove status codes that already have a redirect.
    $selectable_status_codes = array_diff_key($status_codes, $redirects);
    if (!empty($selectable_status_codes)) {
      $form['new_status_code'] = [
        '#type' => 'select',
        '#title' => $this->t('Status code'),
        '#options' => $selectable_status_codes,
        '#description' => $this->t('Select the HTTP status code to add a new redirect for. Selecting "4xx" will apply the redirect as the fallback for all undefined 4xx status codes.'),
      ];
      $form['add'] = [
        '#type' => 'submit',
        '#value' => $this->t('Add'),
        '#name' => 'add',
        '#submit' => ['::addRedirect'],
      ];
    }

    $form['redirects'] = ['#tree' => TRUE];
    foreach ($redirects as $code => $redirect) {
      $form['redirects'][$code] = [
        '#type' => 'fieldset',
        '#title' => $status_codes[$code],
      ];
      $form['redirects'][$code]['redirect_status_code'] = [
        '#type' => 'select',
        '#title' => $this->t('Redirect status'),
        '#options' => $redirect_status_options,
        '#empty_option' => $this->t('- Select -'),
        '#config_target' => "redirect4xx.settings:redirects.{$code}.redirect_status_code",
        '#default_value' => $redirect['redirect_status_code'] ?? '',
        '#description' => $this->t('Select the HTTP status code to use for redirects when a %code error occurs. See <a href=":url" target="_blank">MDN documentation</a> for details.', [
          '%code' => $code,
          ':url' => 'https://developer.mozilla.org/docs/Web/HTTP/Redirections',
        ]),
      ];
      $form['redirects'][$code]['destination_path'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Redirect URL'),
        '#config_target' => "redirect4xx.settings:redirects.{$code}.destination_path",
        '#default_value' => $redirect['destination_path'] ?? '',
        '#description' => $this->t('Enter the destination URL for redirection when a %code error occurs. Examples: <code>/error/%code.html</code></code>.', ['%code' => $code]),
      ];
      $form['redirects'][$code]['enable_on_admin_route'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Enable as an admin route'),
        '#config_target' => "redirect4xx.settings:redirects.{$code}.enable_on_admin_route",
        '#default_value' => $redirect['enable_on_admin_route'] ?? 0,
        '#description' => $this->t('If enabled, this redirect will be handled as an admin route, meaning it will use the admin theme and be subject to admin access controls.'),
      ];
      $form['redirects'][$code]['redirect_authenticated_users'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Redirect authenticated users'),
        '#config_target' => "redirect4xx.settings:redirects.{$code}.redirect_authenticated_users",
        '#default_value' => $redirect['redirect_authenticated_users'] ?? 0,
        '#description' => $this->t('If enabled, authenticated users will also be redirected. If disabled, only anonymous users will be redirected.'),
      ];
      $form['redirects'][$code]['delete'] = [
        '#type' => 'submit',
        '#value' => $this->t('Delete'),
        '#name' => "delete_{$code}",
        '#submit' => ['::deleteRedirect'],
      ];
    }

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);

    $name = $form_state->getTriggeringElement()['#name'];
    if ($name != 'op') {
      return;
    }

    $redirects = $form_state->getValue('redirects') ?? [];
    foreach ($redirects as $code => $redirect) {
      $status_code = $redirect['redirect_status_code'];
      $url = $redirect['destination_path'];

      if ($status_code === '') {
        $form_state->setErrorByName(
          "redirects][$code][status_code",
          $this->t('Please select a redirect status for HTTP %code errors.', ['%code' => $code])
        );
      }

      if ($url === '') {
        $form_state->setErrorByName(
          "redirects][$code][destination_path",
          $this->t('Please enter a valid redirect URL for HTTP %code errors.', ['%code' => $code])
        );
      }
      // Ensure the URL starts with a slash (/) and does not contain a domain.
      elseif (!str_starts_with($url, '/') || UrlHelper::isExternal($url)) {
        $form_state->setErrorByName(
          "redirects][$code][destination_path",
          $this->t('The redirect URL for %code must be an internal path starting with "/".', ['%code' => $code])
        );
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Remove all existing redirect settings before saving the new ones.
    // Individual redirects are saved via `#config_target` property.
    $this->config('redirect4xx.settings')
      ->set('redirects', [])
      ->save();

    // Invalidate the cache tag for 4xx-response.
    \Drupal::service('cache_tags.invalidator')
      ->invalidateTags(['4xx-response']);

    parent::submitForm($form, $form_state);
  }

  /**
   * Displays a warning message indicating that changes require saving.
   *
   * This method adds a warning message to inform the user that the changes
   * will not take effect until they press the save button.
   */
  protected function addSaveReminderMessage() {
    $message = $this->t('This change will not take effect until you press the save button.');
    $this->messenger()->addWarning($message);
  }

  /**
   * Handles adding a new redirect.
   *
   * This method updates the form state by adding a new redirect entry
   * for the selected HTTP status code. The form is then rebuilt to reflect
   * the changes. The redirect is not saved until the form is submitted.
   *
   * @param array $form
   *   The form structure.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function addRedirect(array $form, FormStateInterface $form_state) {
    $new_status_code = $form_state->getValue('new_status_code');
    $redirects = $form_state->getValue('redirects');
    $redirects[$new_status_code] = [];
    $form_state->setValue('redirects', $redirects);
    $form_state->setRebuild();
    $message = $this->t('Redirect added for HTTP @code.', ['@code' => $new_status_code]);
    $this->messenger()->addMessage($message);
    $this->addSaveReminderMessage();
  }

  /**
   * Handles deleting an existing redirect.
   *
   * This method removes the selected redirect from the form state
   * and rebuilds the form to reflect the changes. The deletion is not
   * saved until the form is submitted.
   *
   * @param array $form
   *   The form structure.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function deleteRedirect(array $form, FormStateInterface $form_state) {
    $trigger_element = $form_state->getTriggeringElement();
    $redirect_code = $trigger_element['#parents'][1];
    $redirects = $form_state->getValue('redirects');
    unset($redirects[$redirect_code]);
    $form_state->setValue('redirects', $redirects);
    $form_state->setRebuild();
    $message = $this->t('Redirect deleted for HTTP @code.', ['@code' => $redirect_code]);
    $this->messenger()->addMessage($message);
    $this->addSaveReminderMessage();
  }

}
